from collections import deque
import itertools


def combine_func_join_arrays(array_of_arrays):
    return list(itertools.chain(*array_of_arrays))


class BufferingIterator:
    def __init__(self, it, buffer_size, combine_func=combine_func_join_arrays):
        self._it = iter(it)
        self._buffer_size = buffer_size
        self._combine_func = combine_func

    def __iter__(self):
        return self

    def __next__(self):
        if self._it is None:
            raise StopIteration

        pages = deque()

        for i in range(0, self._buffer_size):
            val = next(self._it, None)
            if val is None:
                self._it = None
                if not pages:
                    # edge case, never had any lines to begin with
                    raise StopIteration
                break
            pages.append(val)

        return self._combine_func(pages)



import requests
import logging


class CursorApiIterator:
    def __init__(self, session: requests.Session,
                         url: str,
                         request_cursor_param_name: str,
                         response_cursor_field_name: str,
                         response_records_field_name: str,
                         extra_params: dict = {},
                         extra_headers: dict = {}):
        self.logger = logging.getLogger(__name__)
        self._session = session
        self._url = url
        self._request_cursor_param_name = request_cursor_param_name
        self._response_cursor_field_name = response_cursor_field_name
        self._response_records_field_name = response_records_field_name
        self._extra_params = extra_params
        self._request_headers = {
            # "Content-type": "application/json",
            "Accept": "application/json",
            **extra_headers
        }
        self._completed = False
        self._next_cursor = None
        self.page_number = 1

    def __iter__(self):
        self.logger.debug(f"Downloading from {self._url}")
        self._completed = False
        self._next_cursor = None
        self.page_number = 1
        return self

    def __next__(self):
        if self._completed:
            raise StopIteration
        params = {**self._extra_params}
        params[self._request_cursor_param_name] = self._next_cursor

        response = self._session.get(self._url, headers=self._request_headers, params=params)
        response.raise_for_status()

        response_json = response.json()

        self._next_cursor = response_json.get(self._response_cursor_field_name)
        self._completed = self._next_cursor is None

        self.logger.debug(
            f"Downloaded page {self.page_number}. Items collected: {len(response_json[self._response_records_field_name])}")
        self.page_number += 1

        return response_json[self._response_records_field_name]





def normalize_with_prototype(prototype, object_to_norm, freestyle_attrs_name="freestyle_attrs"):
    """
    Normalize python dict, so that it has all the fields and only the fields specified in a prototype dict.

    :param prototype: A prototype dict, for example from Swagger doc
    :param object_to_norm: A dict that comes from API
    :param freestyle_attrs_name: it is a name for attribute, when prototype contains a dict like this: "field": {"": ""}
    :return:
    """
    def normalize_with_prototype_rec(prototype, object_to_norm, path):
        if isinstance(prototype, dict):
            normalized_object = {}
            if object_to_norm is None:
                object_to_norm = {}
            if not isinstance(object_to_norm, dict):
                raise Exception(f"{path}: object_to_norm is not dict")
            for k, v in prototype.items():
                if k != "":
                    normalized_object[k] = normalize_with_prototype_rec(prototype[k], object_to_norm.get(k), path=f"{path}.{k}")
                else:
                    normalized_object[freestyle_attrs_name] = [
                        {"key": key, "value": value} for key, value in object_to_norm.items() if key not in prototype.keys() or key == ""
                    ]
            return normalized_object
        elif isinstance(prototype, list):
            if object_to_norm is None:
                object_to_norm = []
            if not isinstance(object_to_norm, list):
                raise Exception(f"{path}: object_to_norm is not list")
            return [normalize_with_prototype_rec(prototype[0], e, path=f"{path}[{idx}]") for idx, e in enumerate(object_to_norm)]
        elif isinstance(prototype, str):
            if object_to_norm is None:
                return None
            if type(prototype) != type(object_to_norm):
                raise Exception(f"{path}: type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
            return object_to_norm
        elif isinstance(prototype, bool):
            if object_to_norm is None:
                return None
            if type(prototype) != type(object_to_norm):
                raise Exception(f"{path}: type(prototype) != type(object_to_norm): {type(prototype)} != {type(object_to_norm)} (of value {object_to_norm})")
            return object_to_norm
        elif isinstance(prototype, int):
            if object_to_norm is None:
                return None
            if not isinstance(prototype, (int, float)):
                raise Exception(f"{path}: expecting int or float, got {type(object_to_norm)} (of value {object_to_norm})")
            if type(object_to_norm) != int:
                object_to_norm = int(object_to_norm)
            return object_to_norm
        elif isinstance(prototype, float):
            if object_to_norm is None:
                return None
            if not isinstance(prototype, (int, float)):
                raise Exception(f"{path}: expecting int or float, got {type(object_to_norm)} (of value {object_to_norm})")
            if type(object_to_norm) != float:
                object_to_norm = float(object_to_norm)
            return object_to_norm
        else:
            raise Exception(f"{path}: prototype data type not supported: {type(prototype)}")
    return normalize_with_prototype_rec(prototype, object_to_norm, "")



import requests
import time
import logging


def prepare_job(session: requests.Session, url: str,
                response_job_id_field_name: str,
                response_state_field_name: str,
                successful_states: [str],
                waiting_states: [str],
                extra_params: dict = {}, extra_headers: dict = {},
                wait_delay_seconds=5):
    logger = logging.getLogger(__name__)
    requestHeaders = {
        "Content-type": "application/json",
        "Accept": "application/json",
        **extra_headers
    }
    response = session.post(url, headers=requestHeaders, json={**extra_params})
    response.raise_for_status()

    job_id = response.json()[response_job_id_field_name]
    state = waiting_states[0]
    state_response_json = None
    while state in waiting_states:
        time.sleep(wait_delay_seconds)
        state_response = session.get(f"{url}/{job_id}", headers=requestHeaders)
        state_response_json = state_response.json()
        logger.debug(state_response_json)
        state = state_response_json[response_state_field_name]
        logger.debug(f"state: {state}")

    if state not in successful_states:
        raise Exception(f"Job failed with state {state}. Response {state_response_json}")

    return job_id




import requests
import logging


class PaginatedApiIterator:
    def __init__(self,
                 session: requests.Session,
                 url: str,
                 request_page_number_param_name: str,
                 # response_page_number_field_name: str,
                 response_page_count_field_name: str,
                 response_records_field_name: str,
                 extra_params: dict = {},
                 extra_headers: dict = {},
                 start_page_number_from_1=True
                 ):
        self.logger = logging.getLogger(__name__)
        self._session = session
        self._url = url
        self._request_page_number_param_name = request_page_number_param_name
        # self._response_page_number_field_name = response_page_number_field_name
        self._response_page_count_field_name = response_page_count_field_name
        self._response_records_field_name = response_records_field_name
        self._extra_params = extra_params
        self._request_headers = {
            # "Content-type": "application/json",
            "Accept": "application/json",
            **extra_headers
        }
        self._start_page_number_from_1 = start_page_number_from_1
        self.page_number = 1 if self._start_page_number_from_1 else 0
        self._completed = False

    def __iter__(self):
        self.logger.debug(f"Downloading from {self._url}")
        self.page_number = 1 if self._start_page_number_from_1 else 0
        self._completed = False
        return self

    def __next__(self):
        if self._completed:
            raise StopIteration

        response = self._session.get(self._url, headers=self._request_headers,
                                     params={self._request_page_number_param_name: self.page_number, **self._extra_params})
        response.raise_for_status()

        response_json = response.json()
        page_count = response_json[self._response_page_count_field_name]

        self.logger.debug(
            f"Downloaded page {self.page_number} out of {page_count}. Items collected: {len(response_json[self._response_records_field_name])}")

        self.page_number += 1

        if (self._start_page_number_from_1 and self.page_number > page_count) or \
                (not self._start_page_number_from_1 and self.page_number >= page_count):
            self._completed = True

        return response_json[self._response_records_field_name]



from .impl.BufferingIterator import *
from .impl.CursorApiIterator import *
from .impl.normalize_dicts import *
from .impl.normalize_with_prototype import *
from .impl.PaginatedApiIterator import *
from .impl.prepare_job import *

__all__ = ["BufferingIterator", "CursorApiIterator", "normalize_dicts", "normalize_with_prototype",
           "PaginatedApiIterator", "prepare_job"]



import json


def normalize_dicts(records_list: list, path: list, separator=".", return_incomplete_records=True, jsonify_lists=False):
    """
    Normalize list of nested python dicts to a list of one-level dicts.

    Example input:
    ```
    data = [
        {
            "id": 1, "name": "John Smith",
            "pets": [
                {"id": 101, "type": "cat", "name": "Kitty", "toys": [{"name": "toy1"}, {"name": "toy2"}]},
                {"id": 102, "type": "dog", "name": "Barky", "toys": [{"name": "toy3"}]}
            ]
        },
        {
            "id": 2, "name": "Sue Smith",
            "pets": [
                {"id": 201, "type": "cat", "name": "Kitten", "toys": [{"name": "toy4"}, {"name": "toy5"}, {"name": "toy6"}]},
                {"id": 202, "type": "dog", "name": "Fury", "toys": []}
            ]
        },
    ]
    ```

    then `normalize_dicts(data, ["pets", "toys"])` would return:

    ```
    [{'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy1'},
     {'id': 1, 'name': 'John Smith', 'pets.id': 101, 'pets.type': 'cat', 'pets.name': 'Kitty', 'pets.toys.name': 'toy2'},
     {'id': 1, 'name': 'John Smith', 'pets.id': 102, 'pets.type': 'dog', 'pets.name': 'Barky', 'pets.toys.name': 'toy3'},
     {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy4'},
     {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy5'},
     {'id': 2, 'name': 'Sue Smith', 'pets.id': 201, 'pets.type': 'cat', 'pets.name': 'Kitten', 'pets.toys.name': 'toy6'},
     {'id': 2, 'name': 'Sue Smith', 'pets.id': 202, 'pets.type': 'dog', 'pets.name': 'Fury'}]
    ```

    :param records_list: list of dicts to be normalized
    :param path: list of str, name of objects in nested dicts to follow when normalizing
    :param separator: str, (defaults to "."), it will be put in names of fields of resulting dicts.
    :param return_incomplete_records: bool (defaults to True), if following the `path` fails because there is no object
        at some level, this flag indicates if that incomplete record should be returned in result list.
    :return: list of normalized dicts.
    """
    def unroll_dict_rec(d: dict, prefix: str = "") -> dict:
        unrolled = {}
        for k, v in d.items():
            if type(v) == dict:
                for u_k, u_v in unroll_dict_rec(v, f"{prefix}{k}{separator}").items():
                    unrolled[u_k] = u_v
            elif jsonify_lists and type(v) == list:
                unrolled[f"{prefix}{k}"] = json.dumps(v)
            else:
                unrolled[f"{prefix}{k}"] = v
        return unrolled

    def normalize_dicts_rec(records_list: list, path: list, prefix: str = "") -> list:
        if records_list is None or len(records_list) == 0:
            records_list = [{}] if return_incomplete_records else []
        if type(records_list) != list:
            raise Exception(f"records_list is not a list at [{prefix}]")
        records_denormalized = []
        for record in records_list:
            if type(record) != dict:
                raise Exception(f"record is not a dict at [{prefix}]")
            keys_to_repeat = list(record.keys())
            if len(path) > 0 and path[0] in keys_to_repeat:
                keys_to_repeat.remove(path[0])
            obj_to_repeat = unroll_dict_rec({k: record[k] for k in keys_to_repeat}, prefix)
            if len(path) > 0:
                records2 = normalize_dicts_rec(record.get(path[0]), path[1:], prefix=f"{prefix}{path[0]}{separator}")
                for rec2 in records2:
                    records_denormalized.append({**obj_to_repeat, **rec2})
            else:
                records_denormalized.append(obj_to_repeat)
        return records_denormalized
    return normalize_dicts_rec(records_list, path)
